package org.apache.harmony.kernel.vm;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2006, 2009  All Rights Reserved
 */

/**
 * This class must be implemented by the vm vendor.
 *
 * Represents the running virtual machine. All VM specific API
 * are implemented on this class.
 * <p>
 * Note that all methods in VM are static. There is no singleton
 * instance which represents the actively running VM.
 */
public final class VM {

	/*
	 * kernelVersion has the format: aabbxxyy
	 * where:
	 * 	aa - major version of kernel. Must equal that stored in jcl.
	 * 	bb - minor version of kernel. Must be >= that in jcl.
	 *	xx - major version of jcl. Must equal that stored in kernel.
	 *	yy - minor version of jcl. Must be >= that in kernel.
	 */
	private static final int kernelVersion = 0x01000100;

/**
 * Prevents this class from being instantiated.
 */
private VM() {
}

/**
 * Returns the ClassLoader of the method (including natives) at the
 * specified depth on the stack of the calling thread. Frames representing
 * the VM implementation of java.lang.reflect are not included in the list.
 *
 * This is not a public method as it can return the bootstrap class
 * loader, which should not be accessed by non-bootstrap classes.
 *
 * Notes: <ul>
 * 	 <li> This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers. </li>
 *
 *	 <li> The item at depth zero is the caller of this method </li>
 * </ul>
 *
 * @param depth the stack depth of the requested ClassLoader
 * @return the ClassLoader at the specified depth
 *
 * @see java.lang.ClassLoader#getStackClassLoader
 */
private static final native ClassLoader getStackClassLoader(int depth);

/**
 * This method must be provided by the vm vendor, as it is used
 * by other provided class implementations. For example,
 * java.io.ObjectInputStream.readObject() and
 * java.io.ObjectInputStream.resolveProxyClass(). It is also useful
 * for other classes, such as java.rmi.server.RMIClassLoader.
 *
 * Walk the stack and answer the most recent non-null and non-bootstrap
 * ClassLoader on the stack of the calling thread. If no such ClassLoader
 * is found, null is returned.
 *
 * Notes:
 * 	 1) This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers.
 *
 * @return the first non-bootstrap ClassLoader on the stack
 */
public static final ClassLoader getNonBootstrapClassLoader() {
	return com.ibm.oti.vm.VM.getNonBootstrapClassLoader();
}

/**
 * Initialize the classloader.
 *
 * @param 		loader ClassLoader
 *					the ClassLoader instance
 * @param 		bootLoader boolean
 *					true for the bootstrap class loader
 */
public static final void initializeClassLoader(ClassLoader loader, boolean bootLoader) {
	com.ibm.oti.vm.VM.initializeClassLoader(loader, bootLoader);
}

/**
 * This method must be included, as it is used by ResourceBundle.getBundle(),
 * and other places as well. The reference implementation of this method uses
 * the getStackClassLoader() method.
 *
 * Returns the ClassLoader of the method that called the caller.
 * i.e. A.x() calls B.y() calls callerClassLoader(),
 * A's ClassLoader will be returned. Returns null for the
 * bootstrap ClassLoader.
 *
 * @return 		a ClassLoader or null for the bootstrap ClassLoader
 *
 * @throws SecurityException when called from a non-bootstrap Class
 */
public static ClassLoader callerClassLoader() {
	ClassLoader loader = getStackClassLoader(2);
	ClassLoader caller = getStackClassLoader(1);
	ClassLoader bootLoader = getStackClassLoader(0);
	if (caller != bootLoader)
		throw new SecurityException();
	if (loader == getStackClassLoader(0)) return null;
	return loader;
}

/**
 * This method must be provided by the vm vendor, as it is used
 * by org.apache.harmony.luni.vm.MsgHelp.setLocale() to get the bootstrap
 * ClassLoader. MsgHelp uses the bootstrap ClassLoader to find the
 * resource bundle of messages packaged with the bootstrap classes.
 * The reference implementation of this method uses the getStackClassLoader()
 * method.
 *
 * Returns the ClassLoader of the method that called the caller.
 * i.e. A.x() calls B.y() calls callerClassLoader(),
 * A's ClassLoader will be returned. Returns null for the
 * bootstrap ClassLoader.
 *
 * @return 		a ClassLoader
 *
 * @throws SecurityException when called from a non-bootstrap Class
 */
public static ClassLoader bootCallerClassLoader() {
	ClassLoader loader = getStackClassLoader(2);
	ClassLoader caller = getStackClassLoader(1);
	ClassLoader bootLoader = getStackClassLoader(0);
	if (caller != bootLoader)
		throw new SecurityException();
	return loader;
}

/**
 * Native used to dump a string to the system console for debugging.
 *
 * @param 		str String
 *					the String to display
 */
public static void dumpString(String str) {
	com.ibm.oti.vm.VM.dumpString(str);
}

/**
 * Returns command line arguments passed to the VM. Internally these are broken into
 * optionString and extraInfo. This only returns the optionString part.
 * <p>
 * @return		a String array containing the optionString part of command line arguments
 */
public static String [] getVMArgs() {
	return com.ibm.oti.vm.VM.getVMArgs();
}

/**
 * This method must be provided by the vm vendor.
 *
 * Returns an int containing the version number of the kernel. Used to check for kernel
 * compatibility.
 *
 * @return an int containing the kernel version number
 */
public static int getKernelVersion() {
	return kernelVersion;
}

/**
 * Answer the class at depth.
 *
 * Notes:
 * 	 1) This method operates on the defining classes of methods on stack.
 *		NOT the classes of receivers.
 *
 *	 2) The item at index zero describes the caller of this method.
 *
 */
public static final native Class getStackClass(int depth);

}

